# This program is free software; you can redistribute it and/or modify it under
# the terms of the (LGPL) GNU Lesser General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library Lesser General Public License
# for more details at ( http://www.gnu.org/licenses/lgpl.html ).
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# written by: Jeff Ortel ( jortel@redhat.com )

"""
XML document reader classes providing integration with the suds library's
caching system.

"""

import suds.cache
import suds.plugin
import suds.sax.parser
import suds.transport

try:
    from hashlib import md5
except ImportError:
    # 'hashlib' package added in Python 2.5 so use the now deprecated/removed
    # 'md5' package in older Python versions.
    from md5 import md5


class Reader(object):
    """
    Provides integration with the cache.

    @ivar options: An options object.
    @type options: I{Options}

    """

    def __init__(self, options):
        """
        @param options: An options object.
        @type options: I{Options}

        """
        self.options = options
        self.plugins = suds.plugin.PluginContainer(options.plugins)

    def mangle(self, name, x):
        """
        Mangle the name by hashing the I{name} and appending I{x}.

        @return: The mangled name.
        @rtype: str

        """
        h = md5(name.encode()).hexdigest()
        return '%s-%s' % (h, x)


class DefinitionsReader(Reader):
    """
    Integrates between the WSDL Definitions object and the object cache.

    @ivar fn: A factory function used to create objects not found in the cache.
    @type fn: I{Constructor}

    """

    def __init__(self, options, fn):
        """
        @param options: An options object.
        @type options: I{Options}
        @param fn: A factory function used to create objects not found in the
            cache.
        @type fn: I{Constructor}

        """
        super(DefinitionsReader, self).__init__(options)
        self.fn = fn

    def open(self, url):
        """
        Open a WSDL schema at the specified I{URL}.

        First, the WSDL schema is looked up in the I{object cache}. If not
        found, a new one constructed using the I{fn} factory function and the
        result is cached for the next open().

        @param url: A WSDL URL.
        @type url: str.
        @return: The WSDL object.
        @rtype: I{Definitions}

        """
        cache = self.__cache()
        id = self.mangle(url, "wsdl")
        wsdl = cache.get(id)
        if wsdl is None:
            wsdl = self.fn(url, self.options)
            cache.put(id, wsdl)
        else:
            # Cached WSDL Definitions objects may have been created with
            # different options so we update them here with our current ones.
            wsdl.options = self.options
            for imp in wsdl.imports:
                imp.imported.options = self.options
        return wsdl

    def __cache(self):
        """
        Get the I{object cache}.

        @return: The I{cache} when I{cachingpolicy} = B{1}.
        @rtype: L{Cache}

        """
        if self.options.cachingpolicy == 1:
            return self.options.cache
        return suds.cache.NoCache()


class DocumentReader(Reader):
    """Integrates between the SAX L{Parser} and the document cache."""

    def open(self, url):
        """
        Open an XML document at the specified I{URL}.

        First, a preparsed document is looked up in the I{object cache}. If not
        found, its content is fetched from an external source and parsed using
        the SAX parser. The result is cached for the next open().

        @param url: A document URL.
        @type url: str.
        @return: The specified XML document.
        @rtype: I{Document}

        """
        cache = self.__cache()
        id = self.mangle(url, "document")
        xml = cache.get(id)
        if xml is None:
            xml = self.__fetch(url)
            cache.put(id, xml)
        self.plugins.document.parsed(url=url, document=xml.root())
        return xml

    def __cache(self):
        """
        Get the I{object cache}.

        @return: The I{cache} when I{cachingpolicy} = B{0}.
        @rtype: L{Cache}

        """
        if self.options.cachingpolicy == 0:
            return self.options.cache
        return suds.cache.NoCache()

    def __fetch(self, url):
        """
        Fetch document content from an external source.

        The document content will first be looked up in the registered document
        store, and if not found there, downloaded using the registered
        transport system.

        Before being returned, the fetched document content first gets
        processed by all the registered 'loaded' plugins.

        @param url: A document URL.
        @type url: str.
        @return: A file pointer to the fetched document content.
        @rtype: file-like

        """
        content = None
        store = self.options.documentStore
        if store is not None:
            content = store.open(url)
        if content is None:
            request = suds.transport.Request(url)
            fp = self.options.transport.open(request)
            try:
                content = fp.read()
            finally:
                fp.close()
        ctx = self.plugins.document.loaded(url=url, document=content)
        content = ctx.document
        sax = suds.sax.parser.Parser()
        return sax.parse(string=content)



# -*- coding: utf-8 -*-

# This program is free software; you can redistribute it and/or modify
# it under the terms of the (LGPL) GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library Lesser General Public License for more details at
# ( http://www.gnu.org/licenses/lgpl.html ).
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# written by: Jurko Gospodnetić ( jurko.gospodnetic@pke.hr )

"""
Suds web service operation invocation function argument parser.

See the parse_args() function description for more detailed information.

"""

__all__ = ["parse_args"]


def parse_args(method_name, param_defs, args, kwargs, external_param_processor,
    extra_parameter_errors):
    """
    Parse arguments for suds web service operation invocation functions.

    Suds prepares Python function objects for invoking web service operations.
    This function implements generic binding agnostic part of processing the
    arguments passed when calling those function objects.

    Argument parsing rules:
      * Each input parameter element should be represented by single regular
        Python function argument.
      * At most one input parameter belonging to a single choice parameter
        structure may have its value specified as something other than None.
      * Positional arguments are mapped to choice group input parameters the
        same as is done for a simple all/sequence group - each in turn.

    Expects to be passed the web service operation's parameter definitions
    (parameter name, type & optional ancestry information) in order and, based
    on that, extracts the values for those parameter from the arguments
    provided in the web service operation invocation call.

    Ancestry information describes parameters constructed based on suds
    library's automatic input parameter structure unwrapping. It is expected to
    include the parameter's XSD schema 'ancestry' context, i.e. a list of all
    the parent XSD schema tags containing the parameter's <element> tag. Such
    ancestry context provides detailed information about how the parameter's
    value is expected to be used, especially in relation to other input
    parameters, e.g. at most one parameter value may be specified for
    parameters directly belonging to the same choice input group.

    Rules on acceptable ancestry items:
      * Ancestry item's choice() method must return whether the item
        represents a <choice> XSD schema tag.
      * Passed ancestry items are used 'by address' internally and the same XSD
        schema tag is expected to be identified by the exact same ancestry item
        object during the whole argument processing.

    During processing, each parameter's definition and value, together with any
    additional pertinent information collected from the encountered parameter
    definition structure, is passed on to the provided external parameter
    processor function. There that information is expected to be used to
    construct the actual binding specific web service operation invocation
    request.

    Raises a TypeError exception in case any argument related errors are
    detected. The exceptions raised have been constructed to make them as
    similar as possible to their respective exceptions raised during regular
    Python function argument checking.

    Does not support multiple same-named input parameters.

    """
    arg_parser = _ArgParser(method_name, param_defs, external_param_processor)
    return arg_parser(args, kwargs, extra_parameter_errors)


class _ArgParser:
    """Internal argument parser implementation function object."""

    def __init__(self, method_name, param_defs, external_param_processor):
        self.__method_name = method_name
        self.__param_defs = param_defs
        self.__external_param_processor = external_param_processor
        self.__stack = []

    def __call__(self, args, kwargs, extra_parameter_errors):
        """
        Runs the main argument parsing operation.

        Passed args & kwargs objects are not modified during parsing.

        Returns an informative 2-tuple containing the number of required &
        allowed arguments.

        """
        assert not self.active(), "recursive argument parsing not allowed"
        self.__init_run(args, kwargs, extra_parameter_errors)
        try:
            self.__process_parameters()
            return self.__all_parameters_processed()
        finally:
            self.__cleanup_run()
            assert not self.active()

    def active(self):
        """
        Return whether this object is currently running argument processing.

        Used to avoid recursively entering argument processing from within an
        external parameter processor.

        """
        return bool(self.__stack)

    def __all_parameters_processed(self):
        """
        Finish the argument processing.

        Should be called after all the web service operation's parameters have
        been successfully processed and, afterwards, no further parameter
        processing is allowed.

        Returns a 2-tuple containing the number of required & allowed
        arguments.

        See the _ArgParser class description for more detailed information.

        """
        assert self.active()
        sentinel_frame = self.__stack[0]
        self.__pop_frames_above(sentinel_frame)
        assert len(self.__stack) == 1
        self.__pop_top_frame()
        assert not self.active()
        args_required = sentinel_frame.args_required()
        args_allowed = sentinel_frame.args_allowed()
        self.__check_for_extra_arguments(args_required, args_allowed)
        return args_required, args_allowed

    def __check_for_extra_arguments(self, args_required, args_allowed):
        """
        Report an error in case any extra arguments are detected.

        Does nothing if reporting extra arguments as exceptions has not been
        enabled.

        May only be called after the argument processing has been completed.

        """
        assert not self.active()
        if not self.__extra_parameter_errors:
            return

        if self.__kwargs:
            param_name = self.__kwargs.keys()[0]
            if param_name in self.__params_with_arguments:
                msg = "got multiple values for parameter '%s'"
            else:
                msg = "got an unexpected keyword argument '%s'"
            self.__error(msg % (param_name,))

        if self.__args:
            def plural_suffix(count):
                if count == 1:
                    return ""
                return "s"
            def plural_was_were(count):
                if count == 1:
                    return "was"
                return "were"
            expected = args_required
            if args_required != args_allowed:
                expected = "%d to %d" % (args_required, args_allowed)
            given = self.__args_count
            msg_parts = ["takes %s positional argument" % (expected,),
                plural_suffix(expected), " but %d " % (given,),
                plural_was_were(given), " given"]
            self.__error("".join(msg_parts))

    def __cleanup_run(self):
        """Cleans up after a completed argument parsing run."""
        self.__stack = []
        assert not self.active()

    def __error(self, message):
        """Report an argument processing error."""
        raise TypeError("%s() %s" % (self.__method_name, message))

    def __frame_factory(self, ancestry_item):
        """Construct a new frame representing the given ancestry item."""
        frame_class = Frame
        if ancestry_item is not None and ancestry_item.choice():
            frame_class = ChoiceFrame
        return frame_class(ancestry_item, self.__error,
            self.__extra_parameter_errors)

    def __get_param_value(self, name):
        """
        Extract a parameter value from the remaining given arguments.

        Returns a 2-tuple consisting of the following:
          * Boolean indicating whether an argument has been specified for the
            requested input parameter.
          * Parameter value.

        """
        if self.__args:
            return True, self.__args.pop(0)
        try:
            value = self.__kwargs.pop(name)
        except KeyError:
            return False, None
        return True, value

    def __in_choice_context(self):
        """
        Whether we are currently processing a choice parameter group.

        This includes processing a parameter defined directly or indirectly
        within such a group.

        May only be called during parameter processing or the result will be
        calculated based on the context left behind by the previous parameter
        processing if any.

        """
        for x in self.__stack:
            if x.__class__ is ChoiceFrame:
                return True
        return False

    def __init_run(self, args, kwargs, extra_parameter_errors):
        """Initializes data for a new argument parsing run."""
        assert not self.active()
        self.__args = list(args)
        self.__kwargs = dict(kwargs)
        self.__extra_parameter_errors = extra_parameter_errors
        self.__args_count = len(args) + len(kwargs)
        self.__params_with_arguments = set()
        self.__stack = []
        self.__push_frame(None)

    def __match_ancestry(self, ancestry):
        """
        Find frames matching the given ancestry.

        Returns a tuple containing the following:
          * Topmost frame matching the given ancestry or the bottom-most sentry
            frame if no frame matches.
          * Unmatched ancestry part.

        """
        stack = self.__stack
        if len(stack) == 1:
            return stack[0], ancestry
        previous = stack[0]
        for frame, n in zip(stack[1:], xrange(len(ancestry))):
            if frame.id() is not ancestry[n]:
                return previous, ancestry[n:]
            previous = frame
        return frame, ancestry[n + 1:]

    def __pop_frames_above(self, frame):
        """Pops all the frames above, but not including the given frame."""
        while self.__stack[-1] is not frame:
            self.__pop_top_frame()
        assert self.__stack

    def __pop_top_frame(self):
        """Pops the top frame off the frame stack."""
        popped = self.__stack.pop()
        if self.__stack:
            self.__stack[-1].process_subframe(popped)

    def __process_parameter(self, param_name, param_type, ancestry=None):
        """Collect values for a given web service operation input parameter."""
        assert self.active()
        param_optional = param_type.optional()
        has_argument, value = self.__get_param_value(param_name)
        if has_argument:
            self.__params_with_arguments.add(param_name)
        self.__update_context(ancestry)
        self.__stack[-1].process_parameter(param_optional, value is not None)
        self.__external_param_processor(param_name, param_type,
            self.__in_choice_context(), value)

    def __process_parameters(self):
        """Collect values for given web service operation input parameters."""
        for pdef in self.__param_defs:
            self.__process_parameter(*pdef)

    def __push_frame(self, ancestry_item):
        """Push a new frame on top of the frame stack."""
        frame = self.__frame_factory(ancestry_item)
        self.__stack.append(frame)

    def __push_frames(self, ancestry):
        """
        Push new frames representing given ancestry items.

        May only be given ancestry items other than None. Ancestry item None
        represents the internal sentinel item and should never appear in a
        given parameter's ancestry information.

        """
        for x in ancestry:
            assert x is not None
            self.__push_frame(x)

    def __update_context(self, ancestry):
        if not ancestry:
            return
        match_result = self.__match_ancestry(ancestry)
        last_matching_frame, unmatched_ancestry = match_result
        self.__pop_frames_above(last_matching_frame)
        self.__push_frames(unmatched_ancestry)


class Frame:
    """
    Base _ArgParser context frame.

    When used directly, as opposed to using a derived class, may represent any
    input parameter context/ancestry item except a choice order indicator.

    """

    def __init__(self, id, error, extra_parameter_errors):
        """
        Construct a new Frame instance.

        Passed error function is used to report any argument checking errors.

        """
        assert self.__class__ != Frame or not id or not id.choice()
        self.__id = id
        self._error = error
        self._extra_parameter_errors = extra_parameter_errors
        self._args_allowed = 0
        self._args_required = 0
        self._has_value = False

    def args_allowed(self):
        return self._args_allowed

    def args_required(self):
        return self._args_required

    def has_value(self):
        return self._has_value

    def id(self):
        return self.__id

    def process_parameter(self, optional, has_value):
        args_required = 1
        if optional:
            args_required = 0
        self._process_item(has_value, 1, args_required)

    def process_subframe(self, subframe):
        self._process_item(
            subframe.has_value(),
            subframe.args_allowed(),
            subframe.args_required())

    def _process_item(self, has_value, args_allowed, args_required):
        self._args_allowed += args_allowed
        self._args_required += args_required
        if has_value:
            self._has_value = True


class ChoiceFrame(Frame):
    """
    _ArgParser context frame representing a choice order indicator.

    A choice requires as many input arguments as are needed to satisfy the
    least requiring of its items. For example, if we use I(n) to identify an
    item requiring n parameter, then a choice containing I(2), I(3) & I(7)
    requires 2 arguments while a choice containing I(5) & I(4) requires 4.

    Accepts an argument for each of its contained elements but allows at most
    one of its directly contained items to have a defined value.

    """

    def __init__(self, id, error, extra_parameter_errors):
        assert id.choice()
        Frame.__init__(self, id, error, extra_parameter_errors)
        self.__has_item = False

    def _process_item(self, has_value, args_allowed, args_required):
        self._args_allowed += args_allowed
        self.__update_args_required_for_item(args_required)
        self.__update_has_value_for_item(has_value)

    def __update_args_required_for_item(self, item_args_required):
        if not self.__has_item:
            self.__has_item = True
            self._args_required = item_args_required
            return
        self._args_required = min(self.args_required(), item_args_required)

    def __update_has_value_for_item(self, item_has_value):
        if item_has_value:
            if self.has_value() and self._extra_parameter_errors:
                self._error("got multiple values for a single choice "
                    "parameter")
            self._has_value = True



# This program is free software; you can redistribute it and/or modify it under
# the terms of the (LGPL) GNU Lesser General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library Lesser General Public License
# for more details at ( http://www.gnu.org/licenses/lgpl.html ).
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# written by: Jeff Ortel ( jortel@redhat.com )

"""
Basic caching classes.

"""

import suds
import suds.sax.element
import suds.sax.parser

import datetime
import os
try:
    import cPickle as pickle
except Exception:
    import pickle
import shutil
import tempfile

from logging import getLogger
log = getLogger(__name__)


class Cache(object):
    """An object cache."""

    def get(self, id):
        """
        Get an object from the cache by id.

        @param id: The object id.
        @type id: str
        @return: The object, else None.
        @rtype: any

        """
        raise Exception("not-implemented")

    def put(self, id, object):
        """
        Put an object into the cache.

        @param id: The object id.
        @type id: str
        @param object: The object to add.
        @type object: any

        """
        raise Exception("not-implemented")

    def purge(self, id):
        """
        Purge an object from the cache by id.

        @param id: A object id.
        @type id: str

        """
        raise Exception("not-implemented")

    def clear(self):
        """Clear all objects from the cache."""
        raise Exception("not-implemented")


class NoCache(Cache):
    """The pass-through object cache."""

    def get(self, id):
        return

    def put(self, id, object):
        pass


class FileCache(Cache):
    """
    A file-based URL cache.

    @cvar fnprefix: The file name prefix.
    @type fnprefix: str
    @cvar remove_default_location_on_exit: Whether to remove the default cache
        location on process exit (default=True).
    @type remove_default_location_on_exit: bool
    @ivar duration: The duration after which cached entries expire (0=never).
    @type duration: datetime.timedelta
    @ivar location: The cached file folder.
    @type location: str

    """
    fnprefix = "suds"
    __default_location = None
    remove_default_location_on_exit = True

    def __init__(self, location=None, **duration):
        """
        Initialized a new FileCache instance.

        If no cache location is specified, a temporary default location will be
        used. Such default cache location will be shared by all FileCache
        instances with no explicitly specified location within the same
        process. The default cache location will be removed automatically on
        process exit unless user sets the remove_default_location_on_exit
        FileCache class attribute to False.

        @param location: The cached file folder.
        @type location: str
        @param duration: The duration after which cached entries expire
            (default: 0=never).
        @type duration: keyword arguments for datetime.timedelta constructor

        """
        if location is None:
            location = self.__get_default_location()
        self.location = location
        self.duration = datetime.timedelta(**duration)
        self.__check_version()

    def clear(self):
        for filename in os.listdir(self.location):
            path = os.path.join(self.location, filename)
            if os.path.isdir(path):
                continue
            if filename.startswith(self.fnprefix):
                os.remove(path)
                log.debug("deleted: %s", path)

    def fnsuffix(self):
        """
        Get the file name suffix.

        @return: The suffix.
        @rtype: str

        """
        return "gcf"

    def get(self, id):
        try:
            f = self._getf(id)
            try:
                return f.read()
            finally:
                f.close()
        except Exception:
            pass

    def purge(self, id):
        filename = self.__filename(id)
        try:
            os.remove(filename)
        except Exception:
            pass

    def put(self, id, data):
        try:
            filename = self.__filename(id)
            f = self.__open(filename, "wb")
            try:
                f.write(data)
            finally:
                f.close()
            return data
        except Exception:
            log.debug(id, exc_info=1)
            return data

    def _getf(self, id):
        """Open a cached file with the given id for reading."""
        try:
            filename = self.__filename(id)
            self.__remove_if_expired(filename)
            return self.__open(filename, "rb")
        except Exception:
            pass

    def __check_version(self):
        path = os.path.join(self.location, "version")
        try:
            f = self.__open(path)
            try:
                version = f.read()
            finally:
                f.close()
            if version != suds.__version__:
                raise Exception()
        except Exception:
            self.clear()
            f = self.__open(path, "w")
            try:
                f.write(suds.__version__)
            finally:
                f.close()

    def __filename(self, id):
        """Return the cache file name for an entry with a given id."""
        suffix = self.fnsuffix()
        filename = "%s-%s.%s" % (self.fnprefix, id, suffix)
        return os.path.join(self.location, filename)

    @staticmethod
    def __get_default_location():
        """
        Returns the current process's default cache location folder.

        The folder is determined lazily on first call.

        """
        if not FileCache.__default_location:
            tmp = tempfile.mkdtemp("suds-default-cache")
            FileCache.__default_location = tmp
            import atexit
            atexit.register(FileCache.__remove_default_location)
        return FileCache.__default_location

    def __mktmp(self):
        """Create the I{location} folder if it does not already exist."""
        try:
            if not os.path.isdir(self.location):
                os.makedirs(self.location)
        except Exception:
            log.debug(self.location, exc_info=1)
        return self

    def __open(self, filename, *args):
        """Open cache file making sure the I{location} folder is created."""
        self.__mktmp()
        return open(filename, *args)

    @staticmethod
    def __remove_default_location():
        """
        Removes the default cache location folder.

        This removal may be disabled by setting the
        remove_default_location_on_exit FileCache class attribute to False.

        """
        if FileCache.remove_default_location_on_exit:
            # We must not load shutil here on-demand as under some
            # circumstances this may cause the shutil.rmtree() operation to
            # fail due to not having some internal module loaded. E.g. this
            # happens if you run the project's test suite using the setup.py
            # test command on Python 2.4.x.
            shutil.rmtree(FileCache.__default_location, ignore_errors=True)

    def __remove_if_expired(self, filename):
        """
        Remove a cached file entry if it expired.

        @param filename: The file name.
        @type filename: str

        """
        if not self.duration:
            return
        created = datetime.datetime.fromtimestamp(os.path.getctime(filename))
        expired = created + self.duration
        if expired < datetime.datetime.now():
            os.remove(filename)
            log.debug("%s expired, deleted", filename)


class DocumentCache(FileCache):
    """XML document file cache."""

    def fnsuffix(self):
        return "xml"

    def get(self, id):
        fp = None
        try:
            fp = self._getf(id)
            if fp is None:
                return None
            p = suds.sax.parser.Parser()
            return p.parse(fp)
        except Exception:
            if fp is not None:
                fp.close()
            self.purge(id)

    def put(self, id, object):
        if isinstance(object,
                (suds.sax.document.Document, suds.sax.element.Element)):
            super(DocumentCache, self).put(id, suds.byte_str(str(object)))
        return object


class ObjectCache(FileCache):
    """
    Pickled object file cache.

    @cvar protocol: The pickling protocol.
    @type protocol: int

    """
    protocol = 2

    def fnsuffix(self):
        return "px"

    def get(self, id):
        fp = None
        try:
            fp = self._getf(id)
            if fp is not None:
                return pickle.load(fp)
        except Exception:
            if fp is not None:
                fp.close()
            self.purge(id)

    def put(self, id, object):
        data = pickle.dumps(object, self.protocol)
        super(ObjectCache, self).put(id, data)
        return object
